/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_pixels.c,v 1.7 1998/09/25 02:17:53 slouken Exp $";
#endif

/* General (mostly internal) pixel/color manipulation routines for SDL */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_sysvideo.h"
#include "SDL_blit.h"
#include "SDL_pixels_c.h"

/* Helper functions */
/*
 * Take some pixel format masks and turn them into pixel format info.
 * Returns 0, or -1 on an out of memory error.
 */
int SDL_ComputePixelFormat(SDL_PixelFormat *format, Uint8 bpp,
			Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask)
{
	Uint32 mask;

	format->BitsPerPixel = bpp;
	format->BytesPerPixel = (bpp+7)/8;
	switch (bpp) {
		case 1:
			/* Create the 2 color black-white palette */
			format->palette = (SDL_Palette *)malloc(
							sizeof(SDL_Palette));
			if ( format->palette == NULL )
				return(-1);
			(format->palette)->ncolors = 2;
			(format->palette)->colors = (SDL_Color *)malloc(
				(format->palette)->ncolors*sizeof(SDL_Color));
			if ( (format->palette)->colors == NULL ) {
				free(format->palette);
				return(-1);
			}
			format->palette->colors[0].r = 0xFF;
			format->palette->colors[0].g = 0xFF;
			format->palette->colors[0].b = 0xFF;
			format->palette->colors[1].r = 0x00;
			format->palette->colors[1].g = 0x00;
			format->palette->colors[1].b = 0x00;
			format->Rloss = 8;
			format->Gloss = 8;
			format->Bloss = 8;
			format->Aloss = 8;
			format->Rshift = 0;
			format->Gshift = 0;
			format->Bshift = 0;
			format->Ashift = 0;
			format->Rmask = 0;
			format->Gmask = 0;
			format->Bmask = 0;
			format->Amask = 0;
			break;

		case 4:
			/* Create the 16 color VGA palette */
			format->palette = (SDL_Palette *)malloc(
							sizeof(SDL_Palette));
			if ( format->palette == NULL )
				return(-1);
			(format->palette)->ncolors = 16;
			(format->palette)->colors = (SDL_Color *)malloc(
				(format->palette)->ncolors*sizeof(SDL_Color));
			if ( (format->palette)->colors == NULL ) {
				free(format->palette);
				return(-1);
			}
			/* Punt for now, will this ever be used? */
			memset((format->palette)->colors, 0,
				(format->palette)->ncolors*sizeof(SDL_Color));

			/* Palettized formats have no mask info */
			format->Rloss = 8;
			format->Gloss = 8;
			format->Bloss = 8;
			format->Aloss = 8;
			format->Rshift = 0;
			format->Gshift = 0;
			format->Bshift = 0;
			format->Ashift = 0;
			format->Rmask = 0;
			format->Gmask = 0;
			format->Bmask = 0;
			format->Amask = 0;
			break;

		case 8:
			/* Create an empty 256 color palette */
			format->palette = (SDL_Palette *)malloc(
							sizeof(SDL_Palette));
			if ( format->palette == NULL )
				return(-1);
			(format->palette)->ncolors = 256;
			(format->palette)->colors = (SDL_Color *)malloc(
				(format->palette)->ncolors*sizeof(SDL_Color));
			if ( (format->palette)->colors == NULL ) {
				free(format->palette);
				return(-1);
			}
			memset((format->palette)->colors, 0,
				(format->palette)->ncolors*sizeof(SDL_Color));

			/* Palettized formats have no mask info */
			format->Rloss = 8;
			format->Gloss = 8;
			format->Bloss = 8;
			format->Aloss = 8;
			format->Rshift = 0;
			format->Gshift = 0;
			format->Bshift = 0;
			format->Ashift = 0;
			format->Rmask = 0;
			format->Gmask = 0;
			format->Bmask = 0;
			format->Amask = 0;
			break;

		default:
			/* No palette, just packed pixel info */
			format->palette = NULL;
			format->Rshift = 0;
			format->Rloss = 8;
			if ( Rmask ) {
				for ( mask = Rmask; !(mask&0x01); mask >>= 1 )
					++format->Rshift;
				for ( ; (mask&0x01); mask >>= 1 )
					--format->Rloss;
			}
			format->Gshift = 0;
			format->Gloss = 8;
			if ( Gmask ) {
				for ( mask = Gmask; !(mask&0x01); mask >>= 1 )
					++format->Gshift;
				for ( ; (mask&0x01); mask >>= 1 )
					--format->Gloss;
			}
			format->Bshift = 0;
			format->Bloss = 8;
			if ( Bmask ) {
				for ( mask = Bmask; !(mask&0x01); mask >>= 1 )
					++format->Bshift;
				for ( ; (mask&0x01); mask >>= 1 )
					--format->Bloss;
			}
			format->Ashift = 0;
			format->Aloss = 8;
			if ( Amask ) {
				for ( mask = Amask; !(mask&0x01); mask >>= 1 )
					++format->Ashift;
				for ( ; (mask&0x01); mask >>= 1 )
					--format->Aloss;
			}
			format->Rmask = Rmask;
			format->Gmask = Gmask;
			format->Bmask = Bmask;
			format->Amask = Amask;
			break;
	}
	/* Calculate some standard bitmasks, if necessary 
	 * Note:  This could conflict with an alpha mask, if given.
	 */
	if ( (bpp > 8) && !format->Rmask && !format->Gmask && !format->Bmask ) {
		/* R-G-B */
		if ( bpp > 24 )
			bpp = 24;
		format->Rloss = 8-(bpp/3);
		format->Gloss = 8-(bpp/3)-(bpp%3);
		format->Bloss = 8-(bpp/3);
		format->Rshift = ((bpp/3)+(bpp%3))+(bpp/3);
		format->Gshift = (bpp/3);
		format->Bshift = 0;
		format->Rmask = ((0xFF>>format->Rloss)<<format->Rshift);
		format->Gmask = ((0xFF>>format->Gloss)<<format->Gshift);
		format->Bmask = ((0xFF>>format->Bloss)<<format->Bshift);
	}
	format->colorkey = 0;
	format->alpha = 0;
	return(0);
}
/*
 * Calculate an 8-bit (3 red, 3 green, 2 blue) dithered palette of colors
 */
void SDL_DitherColors(SDL_Color *colors, int bpp)
{
	Uint8 r, g, b, i;

	/* Calculate the dither values for each index */
	switch (bpp) {
		case 8:
			for ( r=0; r<(1<<3); ++r ) {
				for ( g=0; g<(1<<3); ++g ) {
					for ( b=0; b<(1<<2); ++b ) {
						i = ((r<<(3+2))|(g<<2)|b);
						colors[i].r = r<<(8-3);
						colors[i].g = g<<(8-3);
						colors[i].b = b<<(8-2);
					}
				}
			}
			break;
		default:
			/* ?? */
			break;
	}
}
/* 
 * Calculate the pad-aligned scanline width of a surface
 */
Uint16 SDL_CalculatePitch(SDL_Surface *surface)
{
	Uint16 pitch;

	/* Surface should be 4-byte aligned for speed */
	pitch = surface->w*surface->format->BytesPerPixel;
	switch (surface->format->BitsPerPixel) {
		case 1:
			pitch = (pitch+7)/8;
			break;
		case 4:
			pitch = (pitch+1)/2;
			break;
		default:
			break;
	}
	pitch = (pitch + 3) & ~3;	/* 4-byte aligning */
	return(pitch);
}
/*
 * Match an RGB value to a particular palette index
 */
Uint8 FindColor(SDL_Palette *pal, Uint8 red, Uint8 green, Uint8 blue)
{
	/* Do colorspace distance matching */
	unsigned int smallest;
	unsigned int distance;
	int rd, gd, bd;
	int i;
	Uint8 pixel=0;
		
	smallest = ~0;
	for ( i=0; i<pal->ncolors; ++i ) {
		rd = pal->colors[i].r - red;
		gd = pal->colors[i].g - green;
		bd = pal->colors[i].b - blue;
		distance = (rd*rd)+(gd*gd)+(bd*bd);
		if ( distance < smallest ) {
			pixel = i;
			if ( distance == 0 ) { /* Perfect match! */
				break;
			}
			smallest = distance;
		}
	}
	return(pixel);
}

/* Map from RGB to a 32bit pixel */
Uint32 SDL_MapRGB (SDL_PixelFormat *format, Uint8 r, Uint8 g, Uint8 b)
{
	Uint32 pixel;

	if ( format->palette == NULL ) {
		pixel = ((r>>format->Rloss)<<format->Rshift)|
			((g>>format->Gloss)<<format->Gshift)|
			((b>>format->Bloss)<<format->Bshift);
	} else {
		pixel = (Uint32)FindColor(format->palette, r, g, b);
	}
	return(pixel);
}
void SDL_GetRGB(Uint32 pixel, SDL_PixelFormat *fmt, Uint8 *r,Uint8 *g,Uint8 *b)
{
	if ( fmt->palette == NULL ) {
		*r = (((pixel&fmt->Rmask)>>fmt->Rshift)<<fmt->Rloss);
		*g = (((pixel&fmt->Gmask)>>fmt->Gshift)<<fmt->Gloss);
		*b = (((pixel&fmt->Bmask)>>fmt->Bshift)<<fmt->Bloss);
	} else {
		*r = fmt->palette->colors[pixel].r;
		*g = fmt->palette->colors[pixel].g;
		*b = fmt->palette->colors[pixel].b;
	}
}
/* Map from Palette to Palette */
static Uint8 *Map1to1(SDL_Palette *src, SDL_Palette *dst, int *identical)
{
	Uint8 *map;
	int i;

	if ( identical ) {
		*identical = 0;
		if ( src->ncolors <= dst->ncolors ) {
			for ( i=0; i<src->ncolors; ++i ) {
				if ( (src->colors[i].r != dst->colors[i].r) ||
				     (src->colors[i].g != dst->colors[i].g) ||
				     (src->colors[i].b != dst->colors[i].b) ) {
					break;
				}
			}
			if ( i == src->ncolors ) {
				/* An identical palette, no need to map */
				/* FIXME:
				   We can't distiguish a 1-bit surface from 
				   an 8 bit surface with a 2 color palette.
				   I'm assuming this doesn't happen... ?
				 */
				*identical = 1;
				return(NULL);
			}
		}
	}
	map = (Uint8 *)malloc(src->ncolors);
	if ( map == NULL ) {
		SDL_OutOfMemory();
		return(NULL);
	}
	for ( i=0; i<src->ncolors; ++i ) {
		map[i] = FindColor(dst,
			src->colors[i].r, src->colors[i].g, src->colors[i].b);
	}
	return(map);
}
/* Map from Palette to BitField */
static Uint8 *Map1toN(SDL_Palette *src, SDL_PixelFormat *dst)
{
	Uint8 *map;
	int i;
	Uint8  bpp;

	bpp = ((dst->BytesPerPixel == 3) ? 4 : dst->BytesPerPixel);
	map = (Uint8 *)malloc(src->ncolors*bpp);
	if ( map == NULL ) {
		SDL_OutOfMemory();
		return(NULL);
	}

	/* We memory copy to the pixel map so the endianness is preserved */
	for ( i=0; i<src->ncolors; ++i ) {
		ASSEMBLE_RGB(&map[i*bpp], dst->BytesPerPixel, dst,
			src->colors[i].r, src->colors[i].g, src->colors[i].b);
	}
	return(map);
}
/* Map from BitField to Dithered-Palette to Palette */
static Uint8 *MapNto1(SDL_PixelFormat *src, SDL_Palette *dst, int *identical)
{
	/* Generate a 256 color dither palette */
	SDL_Palette dithered;
	SDL_Color colors[256];

	dithered.ncolors = 256;
	SDL_DitherColors(colors, 8);
	dithered.colors = colors;
	return(Map1to1(&dithered, dst, identical));
}

int SDL_MapSurface (SDL_Surface *surface, SDL_PixelFormat *dstfmt)
{
	SDL_PixelFormat *srcfmt;

	/* Set it up for a new mapping */
	SDL_UnmapSurface(surface);
	if ( dstfmt == NULL ) {	
		SDL_SetError("No destination format");
		return(-1);
	}
	srcfmt = surface->format;

	switch (srcfmt->BytesPerPixel) {
	    case 1:
		switch (dstfmt->BytesPerPixel) {
		    case 1:
			/* Palette --> Palette */
			surface->palmap = Map1to1(srcfmt->palette,
					dstfmt->palette, &surface->identity);
			if ( ! surface->identity ) {
				if ( surface->palmap == NULL ) {
					return(-1);
				}
			}
			if (srcfmt->BitsPerPixel!=dstfmt->BitsPerPixel)
				surface->identity = 0;
			break;

		    default:
			/* Palette --> BitField */
			surface->palmap = Map1toN(srcfmt->palette, dstfmt);
			if ( surface->palmap == NULL ) {
				return(-1);
			}
			break;
		}
		break;
	default:
		switch (dstfmt->BytesPerPixel) {
		    case 1:
			/* BitField --> Palette */
			surface->palmap = MapNto1(srcfmt,
					dstfmt->palette, &surface->identity);
			if ( ! surface->identity ) {
				if ( surface->palmap == NULL ) {
					return(-1);
				}
			}
			surface->identity = 0;	/* Don't optimize to copy */
			break;
		    default:
			/* BitField --> BitField */
			if ( (srcfmt->BytesPerPixel == dstfmt->BytesPerPixel) &&
				     (srcfmt->Rmask == dstfmt->Rmask) && 
				     (srcfmt->Gmask == dstfmt->Gmask) && 
				     (srcfmt->Bmask == dstfmt->Bmask) )
				surface->identity = 1;
			break;
		}
		break;
	}
	/* Choose your blitters wisely */
	SDL_HWCaps.CheckHWAccel(surface);
	surface->dstfmt = dstfmt;
	return(SDL_CalculateBlit(surface));
}

void SDL_UnmapSurface (SDL_Surface *surface)
{
	/* Free any blit mapping */
	surface->identity = 0;
	surface->dstfmt = NULL;
	if ( surface->palmap ) {
		free(surface->palmap);
	}
	surface->palmap = NULL;
}
