/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_surface.c,v 1.8 1998/09/14 06:55:58 slouken Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_pixels_c.h"
#include "SDL_video_c.h"
#include "SDL_sysvideo.h"
#include "SDL_cursor_c.h"
#include "SDL_blit.h"
#include "SDL_RLEaccel_c.h"
#include "SDL_leaks.h"

/* Public routines */
/*
 * Create an empty surface of the appropriate depth
 */
SDL_Surface * SDL_AllocSurface (Uint32 flags,
			int width, int height, int depth,
			Uint32 Rmask, Uint32 Gmask, Uint32 Bmask, Uint32 Amask)
{
	SDL_Surface *surface;

	/* Allocate the surface */
	surface = (SDL_Surface *)malloc(sizeof(*surface));
	if ( surface == NULL ) {
		SDL_OutOfMemory();
		return(NULL);
	}
	surface->flags = SDL_SWSURFACE;
	surface->format = (SDL_PixelFormat *)malloc(sizeof(*(surface->format)));
	if ( surface->format == NULL ) {
		free(surface);
		SDL_OutOfMemory();
		return(NULL);
	}
	if ( SDL_ComputePixelFormat(surface->format, depth,
					Rmask, Gmask, Bmask, Amask) < 0 ) {
		free(surface->format);
		free(surface);
		SDL_OutOfMemory();
		return(NULL);
	}
	if ( Amask ) {
		surface->flags |= SDL_SRCALPHA;
	}
	surface->locked = 0;
	if ( (flags & SDL_THREADSAFE) == SDL_THREADSAFE ) {
		surface->mutex = SDL_CreateMutex();
		if ( surface->mutex != NULL ) {
			surface->flags |= SDL_THREADSAFE;
		}
	} else {
		surface->mutex = NULL;
	}
	surface->w = width;
	surface->h = height;
	surface->pitch = SDL_CalculatePitch(surface);
	surface->pixels = NULL;
	surface->offset = 0;
	SDL_SetClipping(surface, 0, 0, 0, 0);

	/* Allocate the acceleration data */
	surface->identity = 0;
	surface->dstfmt = NULL;
	surface->hw_blit = NULL;
	surface->sw_blit = NULL;
	surface->palmap = NULL;
	surface->hwaccel_data = NULL;
	surface->swaccel_data = (struct private_swdata *)
				malloc(sizeof(*surface->swaccel_data));
	if ( surface->swaccel_data == NULL ) {
		SDL_FreeSurface(surface);
		SDL_OutOfMemory();
		return(NULL);
	}

	/* Get the pixels */
	if ( ((flags&SDL_HWSURFACE) == SDL_SWSURFACE) || 
				(SDL_HWCaps.AllocHWSurface(surface) < 0) ) {
		if ( surface->w && surface->h ) {
			surface->pixels = malloc(surface->h*surface->pitch);
			if ( surface->pixels == NULL ) {
				SDL_FreeSurface(surface);
				SDL_OutOfMemory();
				return(NULL);
			}
			/* This is important for bitmaps */
			memset(surface->pixels, 0, surface->h*surface->pitch);
		}
	}
	surface->refcount = 1;
#ifdef CHECK_LEAKS
	++surfaces_allocated;
#endif
	/* We're ready to roll */
	return(surface);
}
/*
 * Set the color key in a blittable surface
 */
int SDL_SetColorKey (SDL_Surface *surface, Uint32 flag, Uint32 key)
{
	if ( flag ) {
		surface->flags |= SDL_SRCCOLORKEY;
		surface->format->colorkey = key;
		if ( (surface->flags & SDL_HWACCEL) == SDL_HWACCEL ) {
			if ( (SDL_HWCaps.SetHWColorKey == NULL) ||
				(SDL_HWCaps.SetHWColorKey(surface, key) < 0) ) {
				surface->flags &= ~SDL_HWACCEL;
			}
		}
		if ( flag & (SDL_RLEACCEL|SDL_RLEACCELOK) ) {
			surface->flags |= SDL_RLEACCELOK;
		}
		if ( SDL_CalculateBlit(surface) < 0 ) {
			SDL_SetColorKey(surface, 0, 0);
			return(-1);
		}
	} else {
		surface->flags &= ~(SDL_SRCCOLORKEY|SDL_RLEACCELOK);
		surface->format->colorkey = 0;
		SDL_HWCaps.CheckHWAccel(surface);
		SDL_CalculateBlit(surface);
	}
	return(0);
}
int SDL_SetAlpha (SDL_Surface *surface, Uint32 flag, Uint8 value)
{
	if ( flag ) {
		surface->flags |= SDL_SRCALPHA;
		surface->format->alpha = value;
		if ( (surface->flags & SDL_HWACCEL) == SDL_HWACCEL ) {
			if ( (SDL_HWCaps.SetHWAlpha == NULL) ||
				(SDL_HWCaps.SetHWAlpha(surface, value) < 0) ) {
				surface->flags &= ~SDL_HWACCEL;
			}
		}
		if ( SDL_CalculateBlit(surface) < 0 ) {
			SDL_SetAlpha(surface, 0, 0);
			return(-1);
		}
	} else {
		surface->flags &= ~SDL_SRCALPHA;
		surface->format->alpha = 0;
		SDL_HWCaps.CheckHWAccel(surface);
		SDL_CalculateBlit(surface);
	}
	return(0);
}
/*
 * Set the clipping rectangle for a blittable surface
 */
void SDL_SetClipping (SDL_Surface *surface,
			int top, int left, int bottom, int right)
{
	/* Set the clipping rectangle and flags */
	if ( top || left || bottom || right ) {
		surface->clip_minx = left;
		surface->clip_miny = top;
		surface->clip_maxx = right;
		surface->clip_maxy = bottom;
		surface->flags |= SDL_SRCCLIPPING;
	} else {
		surface->clip_minx = 0;
		surface->clip_miny = 0;
		surface->clip_maxx = surface->w-1;
		surface->clip_maxy = surface->h-1;
		surface->flags &= ~SDL_SRCCLIPPING;
	}
}
/* 
 * Set up a blit between two surfaces -- split into three parts:
 * The upper part, SDL_UpperBlit(), performs clipping and rectangle 
 * verification.  The middle part, SDL_MiddleBlit(), then performs
 * thread-safe locking and cursor updating.  The lower part is a 
 * pointer to a low level accelerated blitting function.
 *
 * These parts are separated out and each used internally by this 
 * library in the optimimum places.  They are exported so that if
 * you know exactly what you are doing, you can optimize your code
 * by calling the one(s) you need.
 */
int SDL_MiddleBlit (SDL_Surface *src, SDL_Rect *srcrect,
				SDL_Surface *dst, SDL_Rect *dstrect)
{
	int was_error;
	SDL_blit do_blit;

	/* Figure out which blitter to use */
	if ( (src->flags & SDL_HWACCEL) == SDL_HWACCEL ) {
		do_blit = src->hw_blit;
	} else {
		do_blit = src->sw_blit;
	}
	if ( do_blit == NULL ) {
		SDL_SetError("Surface hasn't been mapped yet");
		return(-1);
	}

	/* Everything is okay at the beginning...  */
	was_error = 0;

	/* Lock the source and destination if they're thread-safe */
	if ( (src->flags & SDL_THREADSAFE) == SDL_THREADSAFE ) {
		if ( SDL_mutexP(src->mutex) < 0 ) {
			was_error = -1;
		} else {
			++src->locked;
		}
	}
	if ( (dst->flags & SDL_THREADSAFE) == SDL_THREADSAFE ) {
		if ( SDL_mutexP(dst->mutex) < 0 ) {
			was_error = -1;
		} else {
			++dst->locked;
		}
	}

	/* Do the blit */
	if ( ! was_error ) {
		/* Update cursor, if necessary */
		if ( SDL_OccludeCursor(dst, dstrect) ) {
			/* Slightly inefficient -- 3 separate update calls */
			SDL_EraseCursor(dst);
			was_error = do_blit(src, srcrect, dst, dstrect);
			SDL_DrawCursor(dst);
		} else {
			was_error = do_blit(src, srcrect, dst, dstrect);
		}
	}

	/* Unlock the source and destination if they're thread-safe */
	if ( src->locked ) {
		SDL_mutexV(src->mutex);
	}
	if ( dst->locked ) {
		SDL_mutexV(dst->mutex);
	}

	/* Blit is done! */
	return(was_error);
}
int SDL_UpperBlit (SDL_Surface *src, SDL_Rect *srcrect,
					SDL_Surface *dst, SDL_Rect *dstrect)
{
	SDL_Rect fullsrc, fulldst;
	int clip_diff;

	/* Set default rectangles */
	if ( dstrect == NULL ) {
		fulldst.x = 0;
		fulldst.y = 0;
		fulldst.w = dst->w;
		fulldst.h = dst->h;
		dstrect = &fulldst;
	}
	if ( srcrect == NULL ) {
		fullsrc.x = 0;
		fullsrc.y = 0;
		fullsrc.w = src->w;
		fullsrc.h = src->h;
		srcrect = &fullsrc;
	}

	/* Verify the destination coordinates */
	if ( dstrect != &fulldst ) {

		/* Check for negative destination coordinates */
		if ( dstrect->x < 0 ) {
			if ( -dstrect->x >= dstrect->w ) {
				srcrect->w = 0;
				dstrect->w = 0;
				return(0);
			}
			srcrect->x -= dstrect->x;
			srcrect->w += dstrect->x;
			dstrect->w += dstrect->x;
			dstrect->x = 0;
		}
		if ( dstrect->y < 0 ) {
			if ( -dstrect->y >= dstrect->h ) {
				srcrect->h = 0;
				dstrect->h = 0;
				return(0);
			}
			srcrect->y -= dstrect->y;
			srcrect->h += dstrect->y;
			dstrect->h += dstrect->y;
			dstrect->y = 0;
		}

		/* Check for overlarge destination rectangle */
		clip_diff = (dstrect->x+dstrect->w)-(dst->w);
		if ( clip_diff > 0 ) {
			if ( clip_diff >= dstrect->w ) {
				srcrect->w = 0;
				dstrect->w = 0;
				return(0);
			}
			dstrect->w -= clip_diff;
			srcrect->w -= clip_diff;
		}
		clip_diff = (dstrect->y+dstrect->h)-(dst->h);
		if ( clip_diff > 0 ) {
			if ( clip_diff >= dstrect->h ) {
				srcrect->h = 0;
				dstrect->h = 0;
				return(0);
			}
			dstrect->h -= clip_diff;
			srcrect->h -= clip_diff;
		}
	}

	/* Verify the source coordinates */
	if ( srcrect != &fullsrc ) {

		/* Check for negative destination coordinates */
		if ( srcrect->x < 0 ) {
			if ( -srcrect->x >= srcrect->w ) {
				srcrect->w = 0;
				dstrect->w = 0;
				return(0);
			}
			srcrect->w += srcrect->x;
			dstrect->w += srcrect->x;
			srcrect->x = 0;
		}
		if ( srcrect->y < 0 ) {
			if ( -srcrect->y >= srcrect->h ) {
				srcrect->h = 0;
				dstrect->h = 0;
				return(0);
			}
			srcrect->h += srcrect->y;
			dstrect->h += srcrect->y;
			srcrect->y = 0;
		}

		/* Check for overlarge source rectangle */
		clip_diff = (srcrect->x+srcrect->w)-(src->w);
		if ( clip_diff > 0 ) {
			if ( clip_diff >= dstrect->w ) {
				srcrect->w = 0;
				dstrect->w = 0;
				return(0);
			}
			dstrect->w -= clip_diff;
			srcrect->w -= clip_diff;
		}
		clip_diff = (srcrect->y+srcrect->h)-(src->h);
		if ( clip_diff > 0 ) {
			if ( clip_diff >= dstrect->h ) {
				srcrect->h = 0;
				dstrect->h = 0;
				return(0);
			}
			dstrect->h -= clip_diff;
			srcrect->h -= clip_diff;
		}
	}

	/* Check for clipping */
	if ( (src->flags & SDL_SRCCLIPPING) == SDL_SRCCLIPPING ) {

		/* Check for clipping and adjust rectangles */
		clip_diff = src->clip_minx-dstrect->x;
		if ( clip_diff > 0 ) {
			if ( clip_diff >= dstrect->w ) {
				srcrect->w = 0;
				dstrect->w = 0;
				return(0);
			}
#ifdef BLIT_DEBUG
 fprintf(stderr, 
	"Blit of %dx%d at (%d,%d) extends %d pixels beyond left edge\n", 
		dstrect->w, dstrect->h, dstrect->x, dstrect->y, clip_diff);
#endif
			dstrect->x += clip_diff;
			dstrect->w -= clip_diff;
			srcrect->x += clip_diff;
			srcrect->w -= clip_diff;
		}
		clip_diff = (dstrect->x+dstrect->w)-src->clip_maxx;
		if ( clip_diff > 0 ) {
			if ( clip_diff >= dstrect->w ) {
				srcrect->w = 0;
				dstrect->w = 0;
				return(0);
			}
#ifdef BLIT_DEBUG
 fprintf(stderr,
	"Blit of %dx%d at (%d,%d) extends %d pixels beyond right edge (%d)\n", 
		dstrect->w, dstrect->h, dstrect->x, dstrect->y, clip_diff,
								src->clip_maxx);
#endif
			dstrect->w -= clip_diff;
			srcrect->w -= clip_diff;
		}
		clip_diff = src->clip_miny-dstrect->y;
		if ( clip_diff > 0 ) {
			if ( clip_diff >= dstrect->h ) {
				srcrect->h = 0;
				dstrect->h = 0;
				return(0);
			}
#ifdef BLIT_DEBUG
 fprintf(stderr,
	"Blit of %dx%d at (%d,%d) extends %d pixels beyond top edge\n", 
		dstrect->w, dstrect->h, dstrect->x, dstrect->y, clip_diff);
#endif
			dstrect->y += clip_diff;
			dstrect->h -= clip_diff;
			srcrect->y += clip_diff;
			srcrect->h -= clip_diff;
		}
		clip_diff = (dstrect->y+dstrect->h)-src->clip_maxy;
		if ( clip_diff > 0 ) {
			if ( clip_diff >= dstrect->h ) {
				srcrect->h = 0;
				dstrect->h = 0;
				return(0);
			}
#ifdef BLIT_DEBUG
 fprintf(stderr,
	"Blit of %dx%d at (%d,%d) extends %d pixels beyond bottom edge\n", 
		dstrect->w, dstrect->h, dstrect->x, dstrect->y, clip_diff);
#endif
			dstrect->h -= clip_diff;
			srcrect->h -= clip_diff;
		}
	}

	/* Check to make sure the blit rectangles are valid */
	if ( (srcrect->w != dstrect->w) || (srcrect->h != dstrect->h) ) {
		SDL_SetError("SDL_UpperBlit: Passed mismatched rectangles");
		return(-1);
	}

#ifdef BLIT_DEBUG
 fprintf(stderr, "UpperBlit %dx%d from 0x%p, (%d,%d) to 0x%p, (%d,%d) ... \n",
					srcrect->w, srcrect->h,
					src->pixels, srcrect->x, srcrect->y,
					dst->pixels, dstrect->x, dstrect->y);
#endif
	/* Do the blit */
	return(SDL_MiddleBlit(src, srcrect, dst, dstrect));
}
/*
 * Lock a surface to directly access the pixels
 * -- Do not call this from any blit function, as SDL_DrawCursor() may recurse
 *    Instead, use:
 *    if ( (surface->flags & SDL_HWSURFACE) == SDL_HWSURFACE )
 *               SDL_HWCaps.LockHWSurface(surface);
 */
int SDL_LockSurface (SDL_Surface *surface)
{
	/* Perform the lock */
	if ( (surface->flags & SDL_THREADSAFE) == SDL_THREADSAFE ) {
		if ( SDL_mutexP(surface->mutex) < 0 ) {
			SDL_SetError("Couldn't lock surface mutex");
			return(-1);
		}
		++surface->locked;
	}
	if ( (surface->flags & SDL_HWSURFACE) == SDL_HWSURFACE ) {
		if ( SDL_HWCaps.LockHWSurface(surface) < 0 ) {
			if ( surface->locked ) {
				--surface->locked;
				SDL_mutexV(surface->mutex);
			}
			return(-1);
		}
	}
	surface->pixels = (Uint8 *)surface->pixels + surface->offset;
	return(0);
}
/*
 * Unlock a previously locked surface
 * -- Do not call this from any blit function, as SDL_DrawCursor() may recurse
 *    Instead, use:
 *    if ( (surface->flags & SDL_HWSURFACE) == SDL_HWSURFACE )
 *               SDL_HWCaps.UnlockHWSurface(surface);
 */
void SDL_UnlockSurface (SDL_Surface *surface)
{
	/* Perform the unlock */
	surface->pixels = (Uint8 *)surface->pixels - surface->offset;

	/* Unlock hardware or accelerated surfaces */
	if ( (surface->flags & SDL_HWSURFACE) == SDL_HWSURFACE ) {
		SDL_HWCaps.UnlockHWSurface(surface);
	} else {
		/* Update RLE encoded surface with new data */
		if ( (surface->flags & SDL_RLEACCEL) == SDL_RLEACCEL ) {
			SDL_RLESurface(surface);
		}
	}

	/* Draw the mouse if necessary */
	if ( (surface == SDL_CursorSurface) && 
				SHOULD_DRAWCURSOR(SDL_cursorstate) ) {
		SDL_DrawCursor(surface);
	}

	/* Release any waiting threads */
	if ( surface->locked ) {
		--surface->locked;
		SDL_mutexV(surface->mutex);
	}
}

/* 
 * Convert a surface into the specified pixel format.
 */
SDL_Surface * SDL_ConvertSurface (SDL_Surface *surface,
					SDL_PixelFormat *format, Uint32 flags)
{
	SDL_Surface *convert;
	SDL_PixelFormat *surface_dstfmt;
	Uint32 colorkey;
	Uint8 alpha;
	Uint32 surface_flags;
	int clip_minx, clip_maxx, clip_miny, clip_maxy;

	/* Check for empty destination palette! (results in empty image) */
	if ( format->BitsPerPixel == 8 ) {
		int i;
		for ( i=0; i<256; ++i ) {
			if ( (format->palette->colors[i].r != 0) ||
			     (format->palette->colors[i].g != 0) ||
			     (format->palette->colors[i].b != 0) )
				break;
		}
		if ( (i == 256) && (surface->format->BitsPerPixel == 8) ) {
			SDL_SetError("Empty destination palette");
			return(NULL);
		}
	}

	/* Create a surface with the format of the video surface */
	convert = SDL_AllocSurface(flags,
				surface->w, surface->h, format->BitsPerPixel,
		format->Rmask, format->Gmask, format->Bmask, format->Amask);
	if ( convert == NULL ) {
		return(NULL);
	}

	/* Copy the palette if any */
	if ( format->palette ) {
		memcpy(convert->format->palette->colors,
				format->palette->colors,
				format->palette->ncolors*sizeof(SDL_Color));
	}

	/* Save the original surface clipping, color key and alpha */
	surface_flags = surface->flags;
	if ( (surface_flags & SDL_SRCCLIPPING) == SDL_SRCCLIPPING ) {
		clip_minx = surface->clip_minx;
		clip_maxx = surface->clip_maxx;
		clip_miny = surface->clip_miny;
		clip_maxy = surface->clip_maxy;
		SDL_SetClipping(surface, 0, 0, 0, 0);
	}
	if ( (surface_flags & SDL_SRCCOLORKEY) == SDL_SRCCOLORKEY ) {
		colorkey = surface->format->colorkey;
		SDL_SetColorKey(surface, 0, 0);
	}
	if ( (surface_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
		alpha = surface->format->alpha;
		SDL_SetAlpha(surface, 0, 0);
	}
	surface_dstfmt = surface->dstfmt;

	/* Copy over the image data */
	if ( (SDL_MapSurface(surface, format) < 0) ||
			(SDL_BlitSurface(surface, NULL, convert, NULL) < 0) ) {
		SDL_FreeSurface(convert);
		convert = NULL;
	}
	SDL_MapSurface(surface, surface_dstfmt);

	/* Clean up the original surface, and update converted surface */
	if ( (surface_flags & SDL_SRCCLIPPING) == SDL_SRCCLIPPING ) {
		if ( convert != NULL ) {
			SDL_SetClipping(convert,
				clip_miny, clip_minx, clip_maxy, clip_maxx);
		}
		SDL_SetClipping(surface,
				clip_miny, clip_minx, clip_maxy, clip_maxx);
	}
	if ( (surface_flags & SDL_SRCCOLORKEY) == SDL_SRCCOLORKEY ) {
		if ( convert != NULL ) {
			Uint8 keyR, keyG, keyB;

			SDL_GetRGB(colorkey,surface->format,&keyR,&keyG,&keyB);
			SDL_SetColorKey(convert, 
				surface_flags&(SDL_SRCCOLORKEY|SDL_RLEACCELOK), 
				SDL_MapRGB(convert->format, keyR, keyG, keyB));
		}
		SDL_SetColorKey(surface, 
				surface_flags&(SDL_SRCCOLORKEY|SDL_RLEACCELOK), 
								colorkey);
	}
	if ( (surface_flags & SDL_SRCALPHA) == SDL_SRCALPHA ) {
		if ( convert != NULL ) {
			SDL_SetAlpha(convert,surface_flags&SDL_SRCALPHA,alpha);
		}
		SDL_SetAlpha(surface, surface_flags&SDL_SRCALPHA, alpha);
	}

	/* We're ready to go! */
	if ( convert ) {
		SDL_MapSurface(convert, format);
	}
	return(convert);
}

/*
 * Free a surface created by the above function.
 */
void SDL_FreeSurface (SDL_Surface *surface)
{
	/* Free anything that's not NULL */
	if ( surface == NULL ) {
		return;
	}
	if ( --surface->refcount > 0 ) {
		return;
	}
	if ( (surface->flags & SDL_RLEACCEL) == SDL_RLEACCEL ) {
		SDL_UnRLESurface(surface);
	}
	if ( surface->hwaccel_data != NULL ) {
		SDL_HWCaps.FreeHWSurface(surface);
	}
	if ( surface->swaccel_data != NULL ) {
		free(surface->swaccel_data);
		surface->swaccel_data = NULL;
	}
	if ( surface->mutex != NULL ) {
		SDL_DestroyMutex(surface->mutex);
		surface->mutex = NULL;
	}
	if ( surface->format ) {
		if ( (surface->format)->palette ) {
			free(((surface->format)->palette)->colors);
			free((surface->format)->palette);
		}
		free(surface->format);
		surface->format = NULL;
	}
	if ( surface->pixels )
		free(surface->pixels);
	SDL_UnmapSurface(surface);
	free(surface);
#ifdef CHECK_LEAKS
	--surfaces_allocated;
#endif
}
