/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_thread.c,v 1.4 1998/08/18 17:22:28 slouken Exp $";
#endif

/* Thread management routines for SDL */

#ifdef FORK_HACK

#include <sys/types.h>
#include <stdlib.h>

#include "SDL_error.h"
#include "SDL_thread.h"

struct SDL_Thread {
	pid_t id;
};

SDL_Thread *SDL_CreateThread(void (*fn)(void *), void *data)
{
	SDL_SetError("Threads are not supported on this platform");
	return(NULL);
}
void SDL_WaitThread(SDL_Thread *thread)
{
}
void SDL_KillThread(SDL_Thread *thread)
{
}

#else

#ifdef SDL_USE_PTHREADS

#include <stdio.h>
#include <signal.h>
#include <pthread.h>

#include "SDL_error.h"
#include "SDL_mutex.h"
#include "SDL_thread.h"


int sig_list[] = {
	SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGALRM, SIGTERM, SIGCHLD, SIGWINCH,
	SIGVTALRM, SIGPROF, 0
};

struct SDL_Thread {
	pthread_t id;
};

typedef struct {
	void (*func)(void *);
	void  *data;
	SDL_mutex *wait;
} thread_args;

static void RunThread(void *data)
{
	thread_args *args;
	void (*userfunc)(void *);
	void  *userdata;
	int i;
	sigset_t mask;

	/* Figure out what function to run */
	args = (thread_args *)data;
	userfunc = args->func;
	userdata = args->data;

	/* Mask asynchronous signals for this thread */
	sigemptyset(&mask);
	for ( i=0; sig_list[i]; ++i ) {
		sigaddset(&mask, sig_list[i]);
	}
	pthread_sigmask(SIG_BLOCK, &mask, NULL);

	/* Allow ourselves to be asynchronously cancelled */
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

	/* Release the parent process */
	SDL_mutexV(args->wait);

	/* Run the function */
	(*userfunc)(userdata);
}

SDL_Thread *SDL_CreateThread(void (*fn)(void *), void *data)
{
	SDL_Thread *thread;
	pthread_attr_t type;
	thread_args *args;
	int ret;

	/* Allocate memory for the thread info structure */
	thread = (SDL_Thread *)malloc(sizeof(*thread));
	if ( thread == NULL ) {
		SDL_OutOfMemory();
		return(NULL);
	}

	/* Set up the arguments for the thread */
	args = (thread_args *)malloc(sizeof(*args));
	if ( args == NULL ) {
		free(thread);
		SDL_OutOfMemory();
		return(NULL);
	}
	args->func = fn;
	args->data = data;
	args->wait = SDL_CreateMutex();
	if ( args->wait == NULL ) {
		free(thread);
		free(args);
		SDL_OutOfMemory();
		return(NULL);
	}
	SDL_mutexP(args->wait);		/* We will block later */

	/* Set the thread attributes */
	if ( pthread_attr_init(&type) != 0 ) {
		free(thread);
		SDL_DestroyMutex(args->wait);
		free(args);
		SDL_SetError("Couldn't initialize pthread attributes");
		return(NULL);
	}
	pthread_attr_setdetachstate(&type, PTHREAD_CREATE_JOINABLE);

	/* Create the thread and go! */
	ret = pthread_create(&thread->id, &type, (void *)RunThread, args);
	pthread_attr_destroy(&type);
	if ( ret >= 0 ) {
		SDL_mutexP(args->wait);
	} else {
		free(thread);
		thread = NULL;
	}
	SDL_DestroyMutex(args->wait);
	free(args);

	/* Everything is running now */
	return(thread);
}

void SDL_WaitThread(SDL_Thread *thread)
{
	if ( thread ) {
		pthread_join(thread->id, NULL);
		free(thread);
	}
}

void SDL_KillThread(SDL_Thread *thread)
{
	if ( thread ) {
		pthread_cancel(thread->id);
		pthread_join(thread->id, NULL);
		free(thread);
	}
}

#else /* Linux-specific clone() based implementation */

#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "SDL_error.h"
#include "SDL_mutex.h"
#include "SDL_thread.h"


/* Stack size for child thread */
#define STACKSIZE 16384*4		/* 16384 is too small */

#ifdef __GLIBC__
#include <schedbits.h>
#else
/* From <linux/sched.h> */
#define CLONE_VM      0x00000100   /* set if VM shared */
#define CLONE_FS      0x00000200   /* set if fs info shared */
#define CLONE_FILES   0x00000400   /* set if open files shared */
#define CLONE_SIGHAND 0x00000800   /* set if signal handlers shared */
#define CLONE_PID     0x00001000   /* set if pid shared */

/* The infamous "start_thread" function, courtesy Linus Torvalds */
extern int clone(int (*fn)(void *arg), void *child_stack, int flags, void *arg);
#endif

int sig_list[] = {
	SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGALRM, SIGTERM, SIGCLD, SIGWINCH,
	SIGVTALRM, SIGPROF, 0
};

struct SDL_Thread {
	pid_t id;
	char *stack;
};

typedef struct {
	void (*func)(void *);
	void  *data;
	SDL_mutex *wait;
} thread_args;

static int RunThread(void *data)
{
	thread_args *args;
	void (*userfunc)(void *);
	void  *userdata;
	int i;
	sigset_t mask;

	/* Figure out what function to run */
	args = (thread_args *)data;
	userfunc = args->func;
	userdata = args->data;

	/* Mask asynchronous signals for this thread */
	sigemptyset(&mask);
	for ( i=0; sig_list[i]; ++i ) {
		sigaddset(&mask, sig_list[i]);
	}
	sigprocmask(SIG_BLOCK, &mask, NULL);

	/* Release the parent process */
	SDL_mutexV(args->wait);

	/* Run the function */
	(*userfunc)(userdata);
	return(0);
}

SDL_Thread *SDL_CreateThread(void (*fn)(void *), void *data)
{
	SDL_Thread *thread;
	thread_args *args;
	int threadpid;
	void **stack;

	/* Allocate memory for the thread info structure */
	thread = (SDL_Thread *)malloc(sizeof(*thread));
	if ( thread == NULL )
		return(NULL);

	/* Allocate memory for thread stack */
	thread->stack = (char *) malloc(STACKSIZE);
	if ( thread->stack == NULL ) {
		free(thread);
		return(NULL);
	}
        stack = (void **) (thread->stack + STACKSIZE);

	/* Set up the arguments for the thread */
	args = (thread_args *)malloc(sizeof(*args));
	if ( args == NULL ) {
		free(thread->stack);
		free(thread);
		return(NULL);
	}
	args->func = fn;
	args->data = data;
	args->wait = SDL_CreateMutex();
	if ( args->wait == NULL ) {
		free(thread->stack);
		free(thread);
		free(args);
		return(NULL);
	}
	SDL_mutexP(args->wait);		/* We will block later */

	/* Create the thread and go! */
	errno = 0;
	threadpid = clone(RunThread, stack,
			(CLONE_VM|CLONE_FS|CLONE_FILES|CLONE_SIGHAND), args);
	if ( threadpid >= 0 ) {
		thread->id = threadpid;
		SDL_mutexP(args->wait);
	} else {
		if ( errno == 0 ) {
			errno = -threadpid;
		}
		free(thread->stack);
		free(thread);
		thread = NULL;
	}
	SDL_DestroyMutex(args->wait);
	free(args);

	/* Everything is running now */
	return(thread);
}

void SDL_WaitThread(SDL_Thread *thread)
{
	if ( thread ) {
#ifdef __WCLONE
		errno = 0;
		while ( errno != ECHILD ) {
			waitpid(thread->id, NULL, __WCLONE);
		}
#else
		/* Ack, ugly ugly hack --
		   wait() doesn't work, waitpid() doesn't work, and ignoring
		   SIG_CHLD doesn't work .. and the child thread is still a 
		   zombie, so kill() doesn't work.
		*/
		char command[1024];

		sprintf(command,
		"ps ax|fgrep -v fgrep|fgrep -v '<zombie>'|fgrep %d >/dev/null",
								thread->id);
		while ( system(command) == 0 )
			sleep(1);
#endif
		free(thread->stack);
		free(thread);
	}
}

void SDL_KillThread(SDL_Thread *thread)
{
	if ( thread ) {
		kill(thread->id, SIGKILL);
		SDL_WaitThread(thread);
	}
}

#endif /* SDL_USE_PTHREADS */

#endif /* FORK_HACK */
