/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_sysvideo.c,v 1.11 1998/10/14 01:14:01 slouken Exp $";
#endif

/* X11 based SDL video driver implementation */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <strings.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#ifdef XFREE86_DGA
#include <X11/extensions/xf86dga.h>
#include <X11/extensions/xf86vmode.h>
#endif

#include "SDL.h"
#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_mouse.h"
#include "SDL_sysvideo.h"
#include "SDL_video_c.h"
#include "SDL_pixels_c.h"
#include "SDL_sysmouse_c.h"
#include "SDL_syswm_c.h"
#include "SDL_lowvideo.h"
#include "blank_cursor.h"


/* Hardware Acceleration functions */
static int SDL_LockHWSurface(SDL_Surface *surface);
static void SDL_UnlockHWSurface(SDL_Surface *surface);

/* Private display data */
Display *SDL_Display = NULL;
Window   SDL_Window = (Window)0;
Colormap SDL_DisplayColormap;
Colormap SDL_PrivateColormap;
Atom     SDL_DELWIN;
WMcursor *SDL_BlankCursor = NULL;
int      X11_key_repeat;
static int XF86_DGA = 0;
static XImage  *SDL_Ximage = NULL;
static GC	SDL_GC;
static Colormap SDL_XColorMap;
static int     *SDL_XPixels;
static SDL_Rect **SDL_modelist = NULL;
static int x11_w, x11_h;

#define X_HIDDEN_SIZE	32		/* starting hidden window size */

/* Shared memory information */
extern int XShmQueryExtension(Display *dpy);	/* Not in X11 headers */
static int use_mitshm;
static XShmSegmentInfo shminfo;

/* Shared memory error handler routine */
static int shm_error;
static int (*X_handler)(Display *, XErrorEvent *);
static int shm_errhandler(Display *d, XErrorEvent *e)
{
        if ( e->error_code == BadAccess ) {
        	++shm_error;
        	return(0);
        } else
		return(X_handler(d,e));
}

/* X11 I/O error handler routine */
static int (*XIO_handler)(Display *);
static int xio_errhandler(Display *d)
{
	/* Ack!  Lost X11 connection! */

	/* Turn on keyboard repeat if necessary */
	if ( X11_key_repeat ) {
		SDL_Display = XOpenDisplay(NULL);
		if ( SDL_Display != NULL ) {
			XAutoRepeatOn(SDL_Display);
			XCloseDisplay(SDL_Display);
		}
	}

	/* We will crash if we try to clean up our display */
	if ( SDL_Ximage ) {
		SDL_VideoSurface->pixels = NULL;
	}
	SDL_Display = NULL;

	/* Continue with the standard X11 error handler */
	return(XIO_handler(d));
}

/* Used to sort video modes largest to smallest */
static int CompareModes(const void *A, const void *B)
{
	const SDL_Rect *a = *(SDL_Rect **)A;
	const SDL_Rect *b = *(SDL_Rect **)B;

	return( (int)(b->w*b->h) - (int)(a->w*a->h) );
}

int SDL_SYS_VideoInit(SDL_PixelFormat *vformat)
{
	int EventBase, ErrorBase;
	int MajorVersion, MinorVersion;
	int flags;
	XGCValues gcv;
	unsigned long Black;

	/* Initialize all variables that we clean on shutdown */
	SDL_Window = (Window)0;
	SDL_Ximage = NULL;
	SDL_XPixels = NULL;
	SDL_PrivateColormap = 0;
	SDL_DisplayColormap = 0;
	SDL_BlankCursor = NULL;
	SDL_modelist = NULL;
	X11_key_repeat = 0;

	/* Open the X11 display */
	SDL_Display = XOpenDisplay(NULL);
	if ( SDL_Display == NULL ) {
		SDL_SetError("Couldn't open X11 display");
		return(-1);
	}

	/* Set the error handler if we lose the X display */
	XIO_handler = XSetIOErrorHandler(xio_errhandler);

	/* Register the "delete window" protocol */
	SDL_DELWIN = XInternAtom(SDL_Display, "WM_DELETE_WINDOW", False);

	/* Determine the screen depth */
	{
		XPixmapFormatValues *pix_format;
		int i, num_formats;

		vformat->BitsPerPixel = DefaultDepth(SDL_Display,
						DefaultScreen(SDL_Display));
		pix_format = XListPixmapFormats(SDL_Display, &num_formats);
		if ( pix_format == NULL ) {
			SDL_SetError("Couldn't determine screen formats");
			return(-1);
		}
		for ( i=0; i<num_formats; ++i ) {
			if ( vformat->BitsPerPixel == pix_format[i].depth )
				break;
		}
		if ( i != num_formats )
			vformat->BitsPerPixel = pix_format[i].bits_per_pixel;
		XFree((char *)pix_format);
	}
	if ( vformat->BitsPerPixel > 8 ) {
		vformat->Rmask = DefaultVisual(SDL_Display,
					DefaultScreen(SDL_Display))->red_mask;
	  	vformat->Gmask = DefaultVisual(SDL_Display,
					DefaultScreen(SDL_Display))->green_mask;
	  	vformat->Bmask = DefaultVisual(SDL_Display,
					DefaultScreen(SDL_Display))->blue_mask;
	}

	/* Check for MIT shared memory extension */
	use_mitshm = XShmQueryExtension(SDL_Display);

#ifdef XFREE86_DGA
	/* Check for XFree96 Direct Graphics Access extension */
	if ( XF86DGAQueryExtension(SDL_Display, &EventBase, &ErrorBase) &&
	     XF86DGAQueryVersion(SDL_Display, &MajorVersion, &MinorVersion) &&
	     XF86DGAQueryDirectVideo(SDL_Display, DefaultScreen(SDL_Display),
				&flags) && (flags & XF86DGADirectPresent) ) {
		int i, nmodes;
		int buggy_metroX;
		XF86VidModeModeInfo **modes;

		XF86_DGA = 1;
		/* Metro-X 4.0 and earlier has a broken implementation of
		   XF86VidModeGetAllModeLines().  When it's fixed, we'll use 
		   VendorRelease(SDL_Display) to figure out if it's buggy.
		 */
		if ( strcmp(ServerVendor(SDL_Display),
					"Metro Link Incorporated") == 0 ) {
			buggy_metroX = 1;
		} else {
			buggy_metroX = 0;
		}
		if ( ! buggy_metroX &&
		     XF86VidModeQueryExtension(SDL_Display,
					&EventBase, &ErrorBase) &&
		     XF86VidModeQueryVersion(SDL_Display,
					&MajorVersion, &MinorVersion) &&
		     (MajorVersion >= 0) && (MinorVersion >= 8) &&
		     XF86VidModeGetAllModeLines(SDL_Display,
				DefaultScreen(SDL_Display), &nmodes, &modes) ) {
			SDL_modelist = (SDL_Rect **)
					malloc((nmodes+1)*sizeof(SDL_Rect *));
			if ( SDL_modelist ) {
				for ( i=0; i<nmodes; ++i ) {
					SDL_modelist[i] = (SDL_Rect *)
						malloc(sizeof(SDL_Rect));
					if ( SDL_modelist[i] == NULL ) {
						SDL_OutOfMemory();
						return(-1);
					}
					SDL_modelist[i]->x = 0;
					SDL_modelist[i]->y = 0;
					SDL_modelist[i]->w = modes[i]->hdisplay;
					SDL_modelist[i]->h = modes[i]->vdisplay;
				}
				SDL_modelist[i] = NULL;

				/* Sort modes largest to smallest */
				qsort(SDL_modelist, nmodes, sizeof(SDL_Rect *),
								CompareModes);
			}
		} else {
			SDL_modelist = (SDL_Rect **)
					malloc((1+1)*sizeof(SDL_Rect *));
			if ( SDL_modelist ) {
				SDL_modelist[0] = (SDL_Rect *)
						malloc(sizeof(SDL_Rect));
				if ( SDL_modelist[0] == NULL ) {
					SDL_OutOfMemory();
					return(-1);
				}
				SDL_modelist[0]->x = 0;
				SDL_modelist[0]->y = 0;
				SDL_modelist[0]->w = WidthOfScreen(
					DefaultScreenOfDisplay(SDL_Display));
				SDL_modelist[0]->h = HeightOfScreen(
					DefaultScreenOfDisplay(SDL_Display));
				SDL_modelist[1] = NULL;
			}
		}
#ifdef XF86_DEBUG
		if ( SDL_modelist ) {
			int i;

			fprintf(stderr, "XF86 mode list:\n");
			for ( i=0; SDL_modelist[i]; ++i ) {
				fprintf(stderr, "\t%dx%d\n",
					SDL_modelist[i]->w, SDL_modelist[i]->h);
			}
		}
#endif
	} else {
		XF86_DGA = 0;
	}
	if ( XF86_DGA ) {
		/* Only root can perform Direct Graphics Access */
		if ( geteuid() != 0 ) {
			XF86_DGA = 0;
		} else
		/* Warn of buffer overflows if this program is set-uid */
		if ( geteuid() != getuid() ) {
			fprintf(stderr,
		"Warning: This set-uid program may be a security risk\n");
		}
	} else {
		/* Draconian measure -- get rid of root priveleges */
		if ( geteuid() == 0 ) {
			setuid(getuid());
		}
	}
#ifdef XF86_DEBUG
	fprintf(stderr, "XFree86 DGA is %s\n", XF86_DGA?"enabled":"disabled");
#endif
#endif /* XFREE86_DGA */

	/* Enumerate the available fullscreen modes */
	/* Create the X11 window */
	Black = BlackPixel(SDL_Display, DefaultScreen(SDL_Display));
	SDL_Window = XCreateSimpleWindow(SDL_Display,
			DefaultRootWindow(SDL_Display), 0, 0,
			X_HIDDEN_SIZE, X_HIDDEN_SIZE, 1, Black, Black);

	/* Allow the window to be deleted by the window manager */
	XSetWMProtocols(SDL_Display, SDL_Window, &SDL_DELWIN, 1);

#define USE_BACKING_STORE		/* Is this a speed penalty? */
#ifdef USE_BACKING_STORE
	/* Cache the window in the server, when possible */
	{
		unsigned long mask;
		Screen *xscreen;
		XSetWindowAttributes a;

		xscreen = DefaultScreenOfDisplay(SDL_Display);
		if ( DoesBackingStore(xscreen) != NotUseful ) {
#ifdef BACKING_DEBUG
  fprintf(stderr, "Server does backing store (%s)\n",
	(DoesBackingStore(xscreen) == Always) ? "always" : "when mapped");
#endif
			mask = CWBackingStore;
			a.backing_store = DoesBackingStore(xscreen);
		} else {
#ifdef BACKING_DEBUG
  fprintf(stderr, "Server does%s save obscured window area\n",
	DoesSaveUnders(xscreen) ? "" : "n't");
#endif
			mask = CWSaveUnder;
			a.save_under = DoesSaveUnders(xscreen);
		}
		XChangeWindowAttributes(SDL_Display, SDL_Window, mask, &a);
	}
#endif /* USE_BACKING_STORE */

	/* Create the graphics context */
	gcv.graphics_exposures = False;
	SDL_GC = XCreateGC(SDL_Display, SDL_Window, GCGraphicsExposures, &gcv);
	if ( ! SDL_GC ) {
		SDL_SetError("Couldn't create graphics context");
		return(-1);
	}

	/* If the surface is palettized, create the colormap */
	if ( vformat->BitsPerPixel == 8 ) {
		int     ncolors;

		/* Get the default display colormap (used by icons) */
		SDL_DisplayColormap = DefaultColormap(SDL_Display,
						DefaultScreen(SDL_Display));
		SDL_XColorMap = SDL_DisplayColormap;

		/* Allocate the pixel flags */
		ncolors = (1<<vformat->BitsPerPixel);
		SDL_XPixels = (int *)malloc(ncolors*sizeof(int));
		if ( SDL_XPixels == NULL ) {
			SDL_OutOfMemory();
			return(-1);
		}
		memset(SDL_XPixels, 0, ncolors*sizeof(*SDL_XPixels));
	}

	/* Create the blank cursor */
	SDL_BlankCursor = SDL_CreateWMCursor(NULL, blank_cdata, blank_cmask,
					BLANK_CWIDTH, BLANK_CHEIGHT,
						BLANK_CHOTX, BLANK_CHOTY);

	/* Set the events that we look for */
	XSelectInput(SDL_Display, SDL_Window,
				( FocusChangeMask | PropertyChangeMask
				| EnterWindowMask | LeaveWindowMask
				| KeyPressMask | KeyReleaseMask
				| ButtonPressMask | ButtonReleaseMask
				| PointerMotionMask
				| ExposureMask | StructureNotifyMask ) );

	/* Fill in some window manager capabilities */
	SDL_WMCaps.SetCaption = SDL_SYS_SetWMCaption;
	SDL_WMCaps.SetIcon = SDL_SYS_SetWMIcon;
	SDL_WMCaps.GetWMInfo = SDL_SYS_GetWMInfo;

	/* Fill in our hardware acceleration capabilities */
	SDL_HWCaps.LockHWSurface = SDL_LockHWSurface;
	SDL_HWCaps.UnlockHWSurface = SDL_UnlockHWSurface;

	/* We're done! */
	return(0);
}

SDL_Rect **SDL_SYS_ListModes(SDL_Surface *screen,
				SDL_PixelFormat *format, Uint32 flags)
{
	if ( format->BitsPerPixel == screen->format->BitsPerPixel ) {
		if ( (flags & SDL_FULLSCREEN) && XF86_DGA ) {
			return(SDL_modelist);
		} else {
			return((SDL_Rect **)-1);
		}
	} else {
		return((SDL_Rect **)0);
	}
}

/* Various screen update functions available */
static void SDL_NormalUpdate(SDL_Surface *screen,int numrects,SDL_Rect *rects);
static void SDL_MITSHMUpdate(SDL_Surface *screen,int numrects,SDL_Rect *rects);
#ifdef XFREE86_DGA
static void SDL_DirectUpdate(SDL_Surface *screen,int numrects,SDL_Rect *rects);
static void SDL_BankedUpdate(SDL_Surface *screen,int numrects,SDL_Rect *rects);
#endif

/* Private structure to keep track of the video card parameters */
struct {
	int active;
	int width;
	int lines;
	char *base;
} vinfo = { 0, 0, 0, NULL };

static int SetupWindow(SDL_Surface *current, Uint32 flags)
{
	/* Allocate shared memory if possible */
	if ( use_mitshm ) {
		shminfo.shmid = shmget(IPC_PRIVATE, current->h*current->pitch,
								IPC_CREAT|0777);
		if ( shminfo.shmid >= 0 ) {
			shminfo.shmaddr = (char *)shmat(shminfo.shmid, 0, 0);
			shminfo.readOnly = False;
			if ( shminfo.shmaddr != (char *)-1 ) {
				shm_error = False;
				X_handler = XSetErrorHandler(shm_errhandler);
				XShmAttach(SDL_Display, &shminfo);
				XSync(SDL_Display, True);
				XSetErrorHandler(X_handler);
				if ( shm_error == True )
					shmdt(shminfo.shmaddr);
			} else {
				shm_error = True;
			}
			shmctl(shminfo.shmid, IPC_RMID, NULL);
		} else {
			shm_error = True;
		}
		if ( shm_error == True )
			use_mitshm = 0;
	}
	if ( use_mitshm ) {
		current->pixels = shminfo.shmaddr;
	} else {
		current->pixels = malloc(current->h*current->pitch);
	}
	if ( current->pixels == NULL ) {
		SDL_OutOfMemory();
		return(-1);
	}

	if ( use_mitshm ) {
		SDL_Ximage = XShmCreateImage(SDL_Display,
			DefaultVisual(SDL_Display, DefaultScreen(SDL_Display)),
			DefaultDepth(SDL_Display, DefaultScreen(SDL_Display)),
			ZPixmap, shminfo.shmaddr, &shminfo, 
						current->w, current->h);
	} else {
		SDL_Ximage = XCreateImage(SDL_Display, 
			DefaultVisual(SDL_Display, DefaultScreen(SDL_Display)),
			DefaultDepth(SDL_Display, DefaultScreen(SDL_Display)),
			ZPixmap, 0, (char *)current->pixels, 
			current->w, current->h,
			((current->format)->BytesPerPixel == 3) ? 32 :
				(current->format)->BytesPerPixel*8, 0);
	}
	if ( SDL_Ximage == NULL ) {
		SDL_SetError("Couldn't create XImage");
		if ( use_mitshm ) {
			XShmDetach(SDL_Display, &shminfo);
			XSync(SDL_Display, False);
			shmdt(shminfo.shmaddr);
			current->pixels = NULL;
		}
		return(-1);
	}

	/* Determine what blit function to use */
	if ( use_mitshm ) {
		SDL_SYS_UpdateRects = SDL_MITSHMUpdate;
	} else {
		SDL_SYS_UpdateRects = SDL_NormalUpdate;
	}
	return(0);
}

static int SetupDirect(SDL_Surface *current, Uint32 flags)
{
#ifdef XFREE86_DGA
	extern void _SDL_Quit(void);
	extern int _SDL_lockSO;

	int i, nmodes;
	int vram, vpage;
	int banked;
	XF86VidModeModeInfo **modes;

	/* Figure out how big the original viewport is */
	if ( ! x11_w || ! x11_h ) {
		XF86DGAGetViewPortSize(SDL_Display, DefaultScreen(SDL_Display),
								&x11_w, &x11_h);
	}

	/* Set the requested video mode */
	XF86VidModeGetAllModeLines(SDL_Display, DefaultScreen(SDL_Display),
							&nmodes, &modes);
	for ( i=0; i<nmodes; ++i ) {
		if ( (current->w == modes[i]->hdisplay) && 
		     (current->h == modes[i]->vdisplay) ) {
			break;
		}
	}
	if ( i == nmodes ) {
		/* This shouldn't happen */
		SDL_SetError("Requested video mode not available");
		return(-1);
	}
	XF86VidModeSwitchToMode(SDL_Display, DefaultScreen(SDL_Display),
								modes[i]);

	/* Get the parameters for the video card */
	XF86DGAGetVideo(SDL_Display, DefaultScreen(SDL_Display),
				&vinfo.base, &vinfo.width, &vpage, &vram);
	vinfo.width *= current->format->BytesPerPixel;
        vinfo.width = (vinfo.width + 3) & ~3;       /* 4-byte aligning */
	vinfo.lines = (vpage/vinfo.width);
	banked = (vinfo.lines < current->h);
#ifdef XF86_DEBUG
	fprintf(stderr,
	"Video card: ram = %dK, pagesize = %d, pitch = %d, base = %p\n",
				vram, vpage, vinfo.width, vinfo.base);
#endif

	/* Arggghh.  The XFree86 3.3.2 implementation of XF86DGAGetVideo()
	   sets atexit() and signal handler functions to the DGA cleanup 
	   function which calls _exit() internally.  This blows away any 
	   atexit() call we might have made previously, and causes a 
	   segmentation fault on exit if we, as a shared library, are 
	   dynamically unloaded.  Hack, hack, hack.
	*/
	atexit(SDL_Quit);
	_SDL_lockSO = 1;

	/* Set direct graphics access */
	XF86DGASetViewPort(SDL_Display, DefaultScreen(SDL_Display), 0, 0);
	XF86DGADirectVideo(SDL_Display, DefaultScreen(SDL_Display),
		XF86DGADirectGraphics|XF86DGADirectMouse|XF86DGADirectKeyb);
	XF86DGASetVidPage(SDL_Display, DefaultScreen(SDL_Display), 0);

	/* Make the input as raw as possible */
	XGrabKeyboard(SDL_Display, SDL_Window, True, GrabModeAsync,
						GrabModeAsync, CurrentTime);
	XGrabPointer(SDL_Display, SDL_Window, True, 
			(ButtonPressMask|ButtonReleaseMask|PointerMotionMask),
			GrabModeAsync, GrabModeAsync, None, None, CurrentTime);

	/* Fill in the current video surface */
	if ( banked || ((flags & SDL_HWSURFACE) == SDL_SWSURFACE) ) {
		current->flags |= SDL_SWSURFACE;
		current->pixels = malloc(current->h*current->pitch);
	} else {
		current->flags |= SDL_HWSURFACE;
		current->pitch = vinfo.width;
		current->pixels = vinfo.base;
	}
	if ( current->pixels == NULL ) {
		SDL_OutOfMemory();
		return(-1);
	}
	current->flags |= SDL_FULLSCREEN;

	/* Determine what blit function to use */
	if ( banked ) {
		SDL_SYS_UpdateRects = SDL_BankedUpdate;
	} else {
		SDL_SYS_UpdateRects = SDL_DirectUpdate;
	}
	vinfo.active = 1;
	return(0);
#else
	SDL_SetError("Logic error: Fullscreen requested without DGA");
	return(-1);
#endif /* XFREE86_DGA */
}

static void ReleaseMode(SDL_Surface *screen)
{
	if ( SDL_Ximage ) {
		XDestroyImage(SDL_Ximage);
		if ( use_mitshm ) {
			XShmDetach(SDL_Display, &shminfo);
			XSync(SDL_Display, False);
			shmdt(shminfo.shmaddr);
		}
		screen->pixels = NULL;
		SDL_Ximage = NULL;
	}
#ifdef XFREE86_DGA
	if ( vinfo.active ) {
		XF86VidModeModeInfo **modes;
		int i, nmodes;

		/* Release the direct input and graphics */
		XUngrabPointer(SDL_Display, CurrentTime);
		XUngrabKeyboard(SDL_Display, CurrentTime);
		XF86DGADirectVideo(SDL_Display, DefaultScreen(SDL_Display), 0);

		/* Set the original video mode */
		XF86VidModeGetAllModeLines(SDL_Display,
				DefaultScreen(SDL_Display), &nmodes, &modes);
		for ( i=0; i<nmodes; ++i ) {
			if ( (x11_w == modes[i]->hdisplay) && 
			     (x11_h == modes[i]->vdisplay) ) {
				break;
			}
		}
		if ( i != nmodes ) {
			XF86VidModeSwitchToMode(SDL_Display,
					DefaultScreen(SDL_Display), modes[i]);
		}
		vinfo.active = 0;
	}
#endif
	/* Turn on keyboard repeat if necessary */
	if ( X11_key_repeat ) {
		XAutoRepeatOn(SDL_Display);
	}
	XUnmapWindow(SDL_Display, SDL_Window);
}


SDL_Surface *SDL_SYS_SetVideoMode(SDL_Surface *current,
			Uint16 width, Uint16 height, Uint8 bpp, Uint32 flags)
{
	XSizeHints   *sizehints;
	XEvent        event;

	/* Clear fullscreen flag if not supported */
	if ( flags & SDL_FULLSCREEN ) {
		if ( ! XF86_DGA ) {
			flags &= ~SDL_FULLSCREEN;
		}
#ifdef XFREE86_DGA
		if ( geteuid() != 0 ) {
			fprintf(stderr,
				"You must be root to set a fullscreen mode\n");
		}
#endif
	}

	/* Release any previously set mode */
	ReleaseMode(current);

	/* Resize the window */
	if ( flags & SDL_FULLSCREEN ) {
		/* Prevent window manager from grabbing us on placement */
		XSetTransientForHint(SDL_Display, SDL_Window, SDL_Window);
	} else {
		/* Allow window manager to grab us and place the window */
		XDeleteProperty(SDL_Display, SDL_Window, XA_WM_TRANSIENT_FOR);
	}
	sizehints = XAllocSizeHints();
	if ( sizehints == NULL ) {
		SDL_SetError("Couldn't allocate size hints");
		return(NULL);
	}
	sizehints->min_width   = width;
	sizehints->min_height  = height;
	sizehints->max_width   = width;
	sizehints->max_height  = height;
	sizehints->flags       = (PMinSize | PMaxSize);
	XSetWMNormalHints(SDL_Display, SDL_Window, sizehints);
	XFree(sizehints);
	XResizeWindow(SDL_Display, SDL_Window, width, height);

	/* Check for keyboard auto-repeat, and turn it off */
	{
		XKeyboardState kbstate;

		XGetKeyboardControl(SDL_Display, &kbstate);
		if ( kbstate.global_auto_repeat ) {
			X11_key_repeat = 1;
		} else {
			X11_key_repeat = 0;
		}
		XAutoRepeatOff(SDL_Display);
	}

	/* Map the window onto the display */
	XMapRaised(SDL_Display, SDL_Window);
	do {
		XNextEvent(SDL_Display, &event);
	} while ( event.type != MapNotify );

	/* Set up the new mode framebuffer */
	current->flags &= SDL_THREADSAFE;	/* Clear flags */
	current->w = width;
	current->h = height;
	current->pitch = SDL_CalculatePitch(current);
	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		if ( SetupDirect(current, flags) == 0 ) {
			flags |= SDL_HWPALETTE;
		} else {
			/* Failed, fall back to window */
			flags &= ~SDL_FULLSCREEN;
		}
	}
	if ( (flags & SDL_FULLSCREEN) != SDL_FULLSCREEN ) {
		if ( SetupWindow(current, flags) < 0 ) {
			return(NULL);
		}
	}

	/* Set the appropriate colormap */
	if ( current->format->palette != NULL ) {
		if ( flags & SDL_HWPALETTE ) {
			current->flags |= SDL_HWPALETTE;
			if ( SDL_PrivateColormap == 0 ) {
				SDL_PrivateColormap = XCreateColormap(SDL_Display, SDL_Window, DefaultVisual(SDL_Display, DefaultScreen(SDL_Display)), AllocAll);
			}
			SDL_XColorMap = SDL_PrivateColormap;
		} else {
			/* Free any currently allocated private colormap */
			if ( SDL_PrivateColormap != 0 ) {
				XFreeColormap(SDL_Display, SDL_PrivateColormap);
				SDL_PrivateColormap = 0;
			}
			SDL_XColorMap = SDL_DisplayColormap;
		}
		XSetWindowColormap(SDL_Display, SDL_Window, SDL_XColorMap);
	}

	/* Make sure that our window has input focus if fullscreen */
	if ( (flags & SDL_FULLSCREEN) == SDL_FULLSCREEN ) {
		XSetInputFocus(SDL_Display, SDL_Window,
					RevertToPointerRoot, CurrentTime);
	}
	XSync(SDL_Display, False);

	/* We're done */
	return(current);
}

/* We don't need to do anything to lock DGA video memory */
static int SDL_LockHWSurface(SDL_Surface *surface)
{
	return(0);
}
static void SDL_UnlockHWSurface(SDL_Surface *surface)
{
	return;
}

static void SDL_NormalUpdate(SDL_Surface *screen, int numrects, SDL_Rect *rects)
{
	int i;

	for ( i=0; i<numrects; ++i ) {
		XPutImage(SDL_Display, SDL_Window, SDL_GC, SDL_Ximage,
				rects[i].x, rects[i].y,
				rects[i].x, rects[i].y, rects[i].w, rects[i].h);
	}
	XSync(SDL_Display, False);
}

static void SDL_MITSHMUpdate(SDL_Surface *screen, int numrects, SDL_Rect *rects)
{
	int i;

	for ( i=0; i<numrects; ++i ) {
		XShmPutImage(SDL_Display, SDL_Window, SDL_GC, SDL_Ximage,
				rects[i].x, rects[i].y,
				rects[i].x, rects[i].y, rects[i].w, rects[i].h,
									False);
	}
	XSync(SDL_Display, False);
}

static void SDL_DirectUpdate(SDL_Surface *screen, int numrects, SDL_Rect *rects)
{
#ifdef XFREE86_DGA
	if ( (screen->flags & SDL_HWSURFACE) == SDL_SWSURFACE ) {
		int i, w, h;
		Uint8 *src, *dst;

		for ( i=0; i<numrects; ++i ) {
			src = (Uint8 *)screen->pixels +
				rects[i].y * screen->pitch +
				rects[i].x * screen->format->BytesPerPixel;
			dst = (Uint8 *)vinfo.base + 
				rects[i].y * vinfo.width +
				rects[i].x * screen->format->BytesPerPixel;
			h = rects[i].h;
			w = rects[i].w * screen->format->BytesPerPixel;
			while ( h-- ) {
				memcpy(dst, src, w);
				src += screen->pitch;
				dst += vinfo.width;
			}
		}
	}
#endif /* XFREE86_DGA */
}

static void SDL_BankedUpdate(SDL_Surface *screen, int numrects, SDL_Rect *rects)
{
#ifdef XFREE86_DGA
	int i, w, h;
	Uint8 *src, *dst;
	int bankleft;
	int bank;
	int xoffset;

	for ( i=0; i<numrects; ++i ) {
		xoffset = rects[i].x * screen->format->BytesPerPixel;
		src = (Uint8 *)screen->pixels +
			rects[i].y * screen->pitch + xoffset;
		dst = (Uint8 *)vinfo.base + 
			(rects[i].y%vinfo.lines) * vinfo.width + xoffset;
		h = rects[i].h;
		w = rects[i].w * screen->format->BytesPerPixel;
		bank = (rects[i].y/vinfo.lines);
		bankleft = vinfo.lines-(rects[i].y%vinfo.lines);
		XF86DGASetVidPage(SDL_Display,DefaultScreen(SDL_Display),bank);
		while ( h-- ) {
			if ( bankleft-- == 0 ) {
				XF86DGASetVidPage(SDL_Display,
					DefaultScreen(SDL_Display), ++bank);
				dst = vinfo.base + xoffset;
				bankleft = vinfo.lines;
			}
			memcpy(dst, src, w);
			src += screen->pitch;
			dst += vinfo.width;
		}
	}
#endif /* XFREE86_DGA */
}

int SDL_SYS_SetColors(SDL_Surface *screen, int firstcolor, int ncolors)
{
	SDL_Color *colors;
	int      alloct_all;
	int      i;
	XColor  *xcmap;

	alloct_all = 1;
	colors = screen->format->palette->colors;
	ncolors = screen->format->palette->ncolors;
	xcmap = (XColor *)malloc(ncolors*sizeof(*xcmap));
	if ( xcmap == NULL ) {
		return(0);
	}

	/* It's easy if we have a private colormap */
	if ( (screen->flags & SDL_HWPALETTE) == SDL_HWPALETTE ) {
		for ( i=0; i<ncolors; ++i ) {
			xcmap[i].pixel = i;
			xcmap[i].red   = (colors[i].r<<8)|colors[i].r;
			xcmap[i].green = (colors[i].g<<8)|colors[i].g;
			xcmap[i].blue  = (colors[i].b<<8)|colors[i].b;
			xcmap[i].flags = (DoRed|DoGreen|DoBlue);
		}
		XStoreColors(SDL_Display, SDL_XColorMap, xcmap, ncolors);
	} else {
		unsigned long pixel;
		XColor        wanted;
		XColor        color;

		/* Free existing allocated colors */
		for ( pixel=0; pixel<ncolors; ++pixel ) {
			while ( SDL_XPixels[pixel] > 0 ) {
				XFreeColors(SDL_Display, SDL_XColorMap, &pixel,
									1, 0);
				--SDL_XPixels[pixel];
			}
		}

		/* Try to allocate all the colors */
		for ( i=0; i<ncolors; ++i ) {
			color.pixel = i;
			color.red   = (colors[i].r<<8);
			color.green = (colors[i].g<<8);
			color.blue  = (colors[i].b<<8);
			color.flags = (DoRed|DoGreen|DoBlue);
			if ( alloct_all ) {
				memcpy(&wanted, &color, sizeof(wanted));
				if (XAllocColor(SDL_Display,
						SDL_XColorMap, &color)) {
					++SDL_XPixels[color.pixel];
				}
				if ( memcmp(&color, &wanted, sizeof(wanted)) )
					alloct_all = 0;
			} else {
				if (XAllocColor(SDL_Display,
						SDL_XColorMap, &color)) {
					++SDL_XPixels[color.pixel];
				}
			}
		}
		if ( ! alloct_all ) {
			/* Copy the colors from the colormap to our palette */
			for ( i=0; i<ncolors; ++i ) {
				xcmap[i].pixel = i;
				xcmap[i].flags = (DoRed|DoGreen|DoBlue);
			}
			XQueryColors(SDL_Display,SDL_XColorMap,xcmap,ncolors);
			for ( i=0; i<ncolors; ++i ) {
				colors[i].r = (xcmap[i].red>>8);
				colors[i].g = (xcmap[i].green>>8);
				colors[i].b = (xcmap[i].blue>>8);
			}
		}
	}
	free(xcmap);
	return(alloct_all);
}

/* Note:  If we are terminated, this could be called in the middle of
   another SDL video routine -- notably SDL_SYS_UpdateRects.
*/
void SDL_SYS_VideoQuit(SDL_Surface *screen)
{
	/* Shutdown everything that's still up */
	if ( SDL_Display != NULL ) {

		/* Start shutting down the windows */
		ReleaseMode(screen);
		if ( SDL_Window ) {
			XDestroyWindow(SDL_Display, SDL_Window);
		}
		if ( SDL_modelist ) {
			int i;

			for ( i=0; SDL_modelist[i]; ++i ) {
				free(SDL_modelist[i]);
			}
			free(SDL_modelist);
			SDL_modelist = NULL;
		}
		if ( SDL_PrivateColormap != 0 ) {
			XFreeColormap(SDL_Display, SDL_PrivateColormap);
			SDL_PrivateColormap = 0;
		}
		if ( SDL_XPixels ) {
			unsigned long pixel;
			int numcolors =
				((screen->format)->palette)->ncolors;
			for ( pixel=0; pixel<numcolors; ++pixel ) {
				while ( SDL_XPixels[pixel] > 0 ) {
					XFreeColors(SDL_Display,
						SDL_DisplayColormap,&pixel,1,0);
					--SDL_XPixels[pixel];
				}
			}
			free(SDL_XPixels);
			SDL_XPixels = NULL;
		} 
		if ( SDL_iconcolors ) {
			unsigned long pixel;
			int numcolors =
				((screen->format)->palette)->ncolors;
			for ( pixel=0; pixel<numcolors; ++pixel ) {
				while ( SDL_iconcolors[pixel] > 0 ) {
					XFreeColors(SDL_Display,
						SDL_DisplayColormap,&pixel,1,0);
					--SDL_iconcolors[pixel];
				}
			}
			free(SDL_iconcolors);
			SDL_iconcolors = NULL;
		} 
		/* Free that blank cursor */
		if ( SDL_BlankCursor != NULL ) {
			SDL_FreeWMCursor(NULL, SDL_BlankCursor);
			SDL_BlankCursor = NULL;
		}

		/* Close the X11 display connection */
		XCloseDisplay(SDL_Display);
		SDL_Display = NULL;
	}
	if ( screen->flags & SDL_HWSURFACE ) {
		/* Direct screen access, no memory buffer */
		screen->pixels = NULL;
	}
}
void SDL_SYS_FinalQuit(void) { }
