/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: load.c,v 1.10 1998/10/12 20:34:51 slouken Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#ifdef __WIN32__
#include <windows.h>
#endif

/* Include definitions of all the datatypes and stub functions */
#include "SDL.h"
#include "SDL_mutex.h"
#include "SDL_thread.h"
#include "SDL_timer.h"
#include "SDL_syswm.h"


/* Define the exported function pointers */
#define DECLARE(type, name, args)	type (*name) args = (void *)0;
#include	"exports.decl"
#include	"sysdep.decl"
#undef DECLARE

/* Define the table from which we load the SDL functions */
#define DECLARE(type, name, args)	{ "_"#name, (void **) &name },
typedef struct {
	const char *symnam;
	void      **symbol;
} symbol_t;

static symbol_t symtab[] = {
#include	"exports.decl"
#include	"sysdep.decl"
  { (char *)0, (void **)0 }
};
#undef DECLARE

/* The dynamic library initialization and finalization functions */
int  (*_init_func)(Uint32 flags) = (void *)0;
void (*_quit_func)(void) = (void *)0;
int  (*_lock_SO)(void) = (void *)0;

/* Local error message handling */
static char *loaderror = (char *)0;
static char *fullerror = (char *)0;
static int   fullerlen = 0;
static void  SetError(char *fmt, const char *a, const char *b)
{
	int newlen;

	/* Allocate space for the error message */
	if ( ! a ) {
		a = "";
	}
	if ( ! b ) {
		b = "";
	}
	newlen = (strlen(fmt)+strlen(a)+strlen(b));
	if ( newlen > fullerlen ) {
		if ( fullerror != (char *)0 ) {
			free(fullerror);
		}
		fullerror = (char *)malloc(newlen);
		if ( fullerror == (char *)0 ) {
			return;
		}
	}
	sprintf(fullerror, fmt, a, b);
}
static char *GetError(void)
{
	return(fullerror);
}
static void  ClrError(void)
{
	if ( fullerror != (char *)0 ) {
		free(fullerror);
		fullerror = (char *)0;
	}
	fullerlen = 0;
}
/* This macro allows errors to propogate in case of library load failure */
#define RESET_ERROR()	(SDL_GetError = GetError)

/* System dependent library loading routines */
static void *libhandle = (void *)0;
#if defined(__unix__)
# include <dlfcn.h>
#elif defined(__WIN32__)
#ifndef NO_DIRECTX_HEADERS
# include <directx.h>
#endif
#elif defined(__BEOS__)
# include <be/kernel/image.h>
#else
#error Unsupported dynamic link environment
#endif /* system type */

static int   SDL_LoadLibrary(void)
{
	char *library;
#if defined(__unix__)
/* * */
	char libstring[BUFSIZ];
	libhandle = NULL;
	/* If no X11 display, try to load GGI library */
	if ( getenv("DISPLAY") == NULL ) {
		sprintf(libstring, "libSDLggi.so.%d.%d", 
					SDL_MAJOR_VERSION, SDL_MINOR_VERSION);
		library = libstring;
		libhandle = dlopen(library, RTLD_NOW);
	}
	/* If we couldn't load library, try default one */
	if ( libhandle == NULL ) {
		sprintf(libstring, "libSDLx11.so.%d.%d", 
					SDL_MAJOR_VERSION, SDL_MINOR_VERSION);
		library = libstring;
		libhandle = dlopen(library, RTLD_NOW);
	}
	loaderror = (char *)dlerror();
#elif defined(__WIN32__)
/* * */
	char errbuf[512];
	HINSTANCE    DDrawDLL;

	/* Clear everything out -- no default library */
	library = (char *)0;
	libhandle = (void *)0;

	/* Version check DDRAW.DLL (Is DirectX okay?) */
	DDrawDLL = LoadLibrary("DDRAW.DLL");
	if ( DDrawDLL != NULL ) {
#ifdef NO_DIRECTX_HEADERS
	  /* Assume DirectX 5.0 or better -- a bad assumption on NT */
	  library = "SDL-dx5.dll";
#else
	  HRESULT (WINAPI *DDrawCreate)(GUID *,LPDIRECTDRAW *,IUnknown *);
	  LPDIRECTDRAW DDraw;

	  /* Try to create a valid DirectDraw object */
	  DDrawCreate = (void *)GetProcAddress(DDrawDLL, "DirectDrawCreate");
	  if ( (DDrawCreate != NULL)
			&& !FAILED(DDrawCreate(NULL, &DDraw, NULL)) ) {
	    if ( !FAILED(IDirectDraw_SetCooperativeLevel(DDraw,
							NULL, DDSCL_NORMAL)) ) {
	      DDSURFACEDESC desc;
	      LPDIRECTDRAWSURFACE  DDrawSurf;
	      LPDIRECTDRAWSURFACE3 DDrawSurf3;

	      /* Try to create a DirectDrawSurface3 object */
	      memset(&desc, 0, sizeof(desc));
	      desc.dwSize = sizeof(desc);
	      desc.dwFlags = DDSD_CAPS;
	      desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	      if ( !FAILED(IDirectDraw_CreateSurface(DDraw, &desc,
							&DDrawSurf, NULL)) ) {
	        if ( !FAILED(IDirectDrawSurface_QueryInterface(DDrawSurf,
			&IID_IDirectDrawSurface3, (LPVOID *)&DDrawSurf3)) ) {
	          /* Yay! */
	          library = "SDL-dx5.dll";

	          /* Clean up.. */
	          IDirectDrawSurface3_Release(DDrawSurf3);
	        }
	        IDirectDrawSurface_Release(DDrawSurf);
	      }
	    }
	    IDirectDraw_Release(DDraw);
	  }
#endif /* NO_DIRECTX_HEADERS */
	  FreeLibrary(DDrawDLL);
	}

	/* If we determined that the DirectX library is okay, load it! */
	if ( library != (char *)0 ) {
		libhandle = (void *)LoadLibrary(library);
	}

	/* Try loading DIB library if DirectX doesn't work */
	if ( libhandle == (void *)0 ) {
		library = "SDL-dib.dll";
		libhandle = (void *)LoadLibrary(library);
	}

	/* Generate an error message if all loads failed */
	if ( libhandle == (void *)0 ) {
		FormatMessage((FORMAT_MESSAGE_IGNORE_INSERTS |
					FORMAT_MESSAGE_FROM_SYSTEM),
				NULL, GetLastError(), 
				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
				errbuf, 512, NULL);
		loaderror = errbuf;
	}
#elif defined(__BEOS__)
/* * */
	char libstring[BUFSIZ];
	static image_id library_id;

	/* Currenly only BWindow is supported */
	sprintf(libstring, "libSDLbwindow.so.%d.%d", 
					SDL_MAJOR_VERSION, SDL_MINOR_VERSION);
	library = libstring;
	library_id = load_add_on(library);
	if ( library_id == B_ERROR ) {
		libhandle = (void *)0;
		loaderror = "BeOS error";
	} else {
		libhandle = (void *)(&library_id);
		loaderror = (char *)0;
	}
#endif /* system type */

	if ( libhandle == (void *)0 ) {
		SetError("Failed loading %s: %s", library, loaderror);
		return(-1);
	}
	return(0);
}
static void *SDL_LoadSym(const char *symnam)
{
	void *symbol;
#if defined(__unix__)
/* * */
	symbol = dlsym(libhandle, symnam);
	loaderror = (char *)dlerror();
#elif defined(__WIN32__)
/* * */
	char errbuf[512];

	symbol = (void *)GetProcAddress((HMODULE)libhandle, symnam);
	if ( symbol == (void *)0 ) {
		FormatMessage((FORMAT_MESSAGE_IGNORE_INSERTS |
					FORMAT_MESSAGE_FROM_SYSTEM),
				NULL, GetLastError(), 
				MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
				errbuf, 512, NULL);
		loaderror = errbuf;
	}
#elif defined(__BEOS__)
/* * */
	if ( get_image_symbol(*((image_id *)libhandle),
		symnam, B_SYMBOL_TYPE_TEXT, &symbol) != B_NO_ERROR ) {
		symbol = (void *)0;
		loaderror = "Symbol not found";
	} else {
		loaderror = (char *)0;
	}
#endif /* system type */

	if ( symbol == (void *)0 ) {
		SetError("Failed loading %s: %s", symnam, loaderror);
	}
	return(symbol);
}
static void  SDL_UnloadLibrary(void)
{
	/* Don't do anything if SDL isn't loaded */
	if ( libhandle == (void *)0 ) {
		return;
	}

	/* Unload the SDL shared library */
#if defined(__unix__)
/* * */
	dlclose(libhandle);
#elif defined(__WIN32__)
/* * */
	FreeLibrary((HMODULE)libhandle);
#elif defined(__BEOS__)
/* * */
	unload_add_on(*((image_id *)libhandle));
#endif /* system type */

	libhandle = (void *)0;
}

static void SDL_Unload_SO(void)
{
	int i;

	_init_func = (void *)0;
	for ( i=0; symtab[i].symnam; ++i ) {
		*symtab[i].symbol = (void *)0;
	}
	RESET_ERROR();
	if ( (_lock_SO == (void *)0) || ! (*_lock_SO)() ) {
		SDL_UnloadLibrary();
	}
}

static int SDL_Load_SO(void)
{
	int i;

	/* Load the SDL shared library */
	RESET_ERROR();
	if ( SDL_LoadLibrary() < 0 ) {
		return(-1);
	}
#ifdef LOAD_DEBUG
	fprintf(stderr, "Loaded SDL library\n");
#endif
	/* Load the initialization code */
	_init_func = SDL_LoadSym("_SDL_Init");
	_quit_func = SDL_LoadSym("_SDL_Quit");
	_lock_SO = SDL_LoadSym("_SDL_LockSO");
	if ( (_init_func == (void *)0) || (_quit_func == (void *)0) ) {
		SDL_Unload_SO();
		return(-1);
	}

	/* Load all of the function pointers */
	for ( i=0; symtab[i].symnam; ++i ) {
		*symtab[i].symbol = SDL_LoadSym(symtab[i].symnam);
		if ( *symtab[i].symbol == (void *)0 ) {
			SDL_Unload_SO();
			return(-1);
		}
	}
#ifdef LOAD_DEBUG
	fprintf(stderr, "Loaded SDL functions\n");
#endif
	/* (leak memory only on error) */
	ClrError();
	return(0);
}

/* Stubs that load all of our symbols and initialize the SDL library */
int SDL_Init(Uint32 flags)
{
	/* Check to make sure the library isn't already loaded */
	if ( _init_func == (void *)0 ) {
		if ( SDL_Load_SO() < 0 ) {
			return(-1);
		}
		/* Check the library version */
		{ const SDL_version *lib_version = SDL_Linked_Version();
		  SDL_version our_version;

		  SDL_VERSION(&our_version);
		  if ( (our_version.major != lib_version->major) ||
		       (our_version.minor != lib_version->minor) ||
			(((our_version.minor%2) != 0) &&
		         (our_version.patch != lib_version->patch)) ) {
			fprintf(stderr,
"Warning: SDL version mismatch -- expected: %d.%d.%d, loaded: %d.%d.%d\n",
				our_version.major, our_version.minor,
							our_version.patch,
				lib_version->major, lib_version->minor,
							lib_version->patch);
		  }
		}
	}
#ifdef LOAD_DEBUG
	fprintf(stderr, "Running init function: 0x%p\n", _init_func);
#endif
	return((*_init_func)(flags));
}

void SDL_Quit(void)
{
	if ( _init_func != (void *)0 ) {
		(*_quit_func)();
		SDL_Unload_SO();
	}
}
